#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
XPathToken class and derived classes for other XPath objects (functions, constructors,
axes, maps, arrays). XPath's error creation and node helper functions are embedded in
XPathToken class, in order to raise errors related to token instances.
"""
from dataclasses import dataclass
from typing import Any

from elementpath.datatypes import UntypedAtomic, QName

from .base import XPathToken
from .axes import XPathAxis
from .functions import XPathFunction, ExternalFunction
from .contructors import XPathConstructor, SchemaConstructor
from .maps import XPathMap
from .arrays import XPathArray
from .tokens import ValueToken, ProxyToken, NameToken, PrefixedNameToken, \
    BracedNameToken, VariableToken, AsteriskToken, ParentShortcutToken, \
    ContextItemToken

__all__ = ['XPathToken', 'XPathAxis', 'XPathFunction', 'XPathConstructor',
           'XPathMap', 'XPathArray', 'ValueToken', 'ProxyToken', 'NameToken',
           'PrefixedNameToken', 'BracedNameToken', 'VariableToken', 'AsteriskToken',
           'ParentShortcutToken', 'ContextItemToken', 'SchemaConstructor',
           'ExternalFunction', 'TokenRegistry']


@dataclass(frozen=True, slots=True)
class TokenRegistry:
    """A registry of classes, helpers and instances used commonly by XPath tokens."""

    # Token base classes
    base_token: type[XPathToken] = XPathToken
    axis_token: type[XPathAxis] = XPathAxis
    function_token: type[XPathFunction] = XPathFunction
    external_function_token: type[ExternalFunction] = ExternalFunction
    constructor_token: type[XPathConstructor] = XPathConstructor
    schema_constructor_token: type[SchemaConstructor] = SchemaConstructor
    array_token: type[XPathArray] = XPathArray
    map_token: type[XPathMap] = XPathMap

    # Other token common classes
    value_token: type[ValueToken] = ValueToken
    proxy_token: type[ProxyToken] = ProxyToken
    name_token: type[NameToken] = NameToken
    prefixed_name_token: type[PrefixedNameToken] = PrefixedNameToken
    braced_name_token: type[BracedNameToken] = BracedNameToken
    variable_token: type[VariableToken] = VariableToken
    asterisk_token: type[AsteriskToken] = AsteriskToken
    parent_shortcut_token: type[ParentShortcutToken] = ParentShortcutToken
    context_item_token: type[ContextItemToken] = ContextItemToken

    QName: type[QName] = QName
    UntypedAtomic: type[UntypedAtomic] = UntypedAtomic

    class __Name:
        name: str | None = None

    def __set_name__(self, owner: type[Any], name: str) -> None:
        self.__Name.name = name

    def __set__(self, instance: Any, value: Any) -> None:
        raise AttributeError("Can't set attribute {!r}".format(self.__Name.name))

    def __delete__(self, instance: Any) -> None:
        raise AttributeError("Can't delete attribute {!r}".format(self.__Name.name))


XPathToken.registry = TokenRegistry()



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from collections.abc import Iterator

import elementpath.aliases as ta
from elementpath.xpath_context import XPathContext
from .base import XPathToken


class XPathAxis(XPathToken):
    pattern = r'\b[^\d\W][\w.\-\xb7\u0300-\u036F\u203F\u2040]*(?=\s*\:\:|\s*\(\:.*\:\)\s*\:\:)'
    label = 'axis'
    reverse_axis: bool = False

    def __str__(self) -> str:
        return f'{self.symbol!r} axis'

    def nud(self) -> 'XPathAxis':
        self.parser.advance('::')
        self.parser.expected_next(
            '(name)', '*', '{', 'Q{', 'text', 'node', 'document-node',
            'comment', 'processing-instruction', 'element', 'attribute',
            'schema-attribute', 'schema-element', 'namespace-node',
        )
        self._items[:] = self.parser.expression(rbp=self.rbp),
        return self

    @property
    def source(self) -> str:
        return '%s::%s' % (self.symbol, self[0].source)

    def select_with_focus(self, context: XPathContext) -> Iterator[ta.ItemType]:
        """Select item with an inner focus on dynamic context."""
        status = context.item, context.size, context.position, context.axis
        results = [x for x in self.select(context)]
        context.item, context.size, context.position, context.axis = status
        context.axis = None

        if self.reverse_axis:
            context.size = context.position = len(results)
            for context.item in results:
                yield context.item
                context.position -= 1
        else:
            context.size = len(results)
            for context.position, context.item in enumerate(results, start=1):
                yield context.item

        context.item, context.size, context.position, context.axis = status



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from collections.abc import Iterable, Iterator
from typing import Any, Optional

import elementpath.aliases as ta

from elementpath.exceptions import ElementPathValueError
from elementpath.sequences import xlist, XSequence
from elementpath.helpers import split_function_test

from elementpath.sequence_types import match_sequence_type
from .functions import XPathFunction


class XPathArray(XPathFunction):
    """
    A token for processing XPath 3.1+ arrays.
    """
    symbol = 'array'
    label = 'array'
    pattern = r'(?<!\$)\barray(?=\s*(?:\(\:.*\:\))?\s*\{(?!\:))'
    _array: Optional[list[ta.ValueType]] = None

    def __init__(self, parser: ta.XPathParserType,
                 items: Optional[Iterable[Any]] = None) -> None:
        if items is not None:
            self._array = [
                x if not isinstance(x, (tuple, list, XSequence)) else xlist(x) for x in items
            ]
        super().__init__(parser)

    def __repr__(self) -> str:
        if self._array is not None:
            return f'<{self.__class__.__name__} object at {hex(id(self))}>'
        return "<{} object (not evaluated constructor) at {}>".format(
            self.__class__.__name__, hex(id(self))
        )

    def __str__(self) -> str:
        if self._array is not None:
            return f'[{", ".join(map(repr, self._array))}]'

        items_desc = f'{len(self)} items' if len(self) != 1 else '1 item'
        if self.symbol == 'array':
            return f'not evaluated curly array constructor with {items_desc}'
        return f'not evaluated square array constructor with {items_desc}'

    def __len__(self) -> int:
        if self._array is None:
            return len(self._items)
        return len(self._array)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, XPathArray):
            if self._array is None or other._array is None:
                raise ElementPathValueError("cannot compare not evaluated arrays")
            return self._array == other._array
        return NotImplemented

    @property
    def source(self) -> str:
        if self._array is None:
            items = ', '.join(f'{tk.source}' for tk in self)
        else:
            items = ', '.join(f'{v!r}' for v in self._array)
        return f'array{{{items}}}' if self.symbol == 'array' else f'[{items}]'

    def nud(self) -> 'XPathArray':
        self.parser.advance('{')
        del self._items[:]
        if self.parser.next_token.symbol not in ('}', '(end)'):
            while True:
                self._items.append(self.parser.expression(5))
                if self.parser.next_token.symbol != ',':
                    break
                self.parser.advance()

        self.parser.advance('}')
        return self

    def evaluate(self, context: ta.ContextType = None) -> 'XPathArray':
        if self._array is not None:
            return self
        return XPathArray(self.parser, items=self._evaluate(context))

    def _evaluate(self, context: ta.ContextType = None) -> list[ta.ValueType]:
        if self.symbol == 'array':
            # A comma in a curly array constructor is the comma operator, not a delimiter.
            items: list[ta.ValueType] = []
            for tk in self._items:
                items.extend(tk.select(context))
            return items
        else:
            return [tk.evaluate(context) for tk in self._items]

    def __call__(self, *args: ta.FunctionArgType, context: ta.ContextType = None) -> ta.ValueType:
        if len(args) != 1 or not isinstance(args[0], int):
            raise self.error('XPTY0004', 'exactly one xs:integer argument is expected')

        position = args[0]
        if position <= 0:
            raise self.error('FOAY0001')

        if self._array is not None:
            items = self._array
        else:
            items = self._evaluate(context)

        try:
            return items[position - 1]
        except IndexError:
            raise self.error('FOAY0001')

    def items(self, context: ta.ContextType = None) -> list[ta.ValueType]:
        if self._array is not None:
            return self._array
        return self._evaluate(context)

    def iter_flatten(self, context: ta.ContextType = None) -> Iterator[ta.ItemType]:
        if self._array is not None:
            items = self._array
        else:
            items = self._evaluate(context)

        for item in items:
            if isinstance(item, XPathArray):
                yield from item.iter_flatten(context)
            elif isinstance(item, list):
                yield from item
            else:
                yield item

    def match_function_test(self, function_test: ta.SequenceTypesType,
                            as_argument: bool = False) -> bool:
        if isinstance(function_test, (list, tuple)):
            sequence_types = function_test
        else:
            sequence_types = split_function_test(function_test)

        if not sequence_types or not sequence_types[-1]:
            return False
        elif sequence_types[0] == '*':
            return True
        elif len(sequence_types) != 2:
            return False

        index_type, value_type = sequence_types
        if index_type.endswith(('+', '*')):
            return False

        return match_sequence_type(1, index_type) and \
            all(match_sequence_type(v, value_type, self.parser) for v in self.items())



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from collections.abc import Callable, Iterator
from copy import copy
from typing import Optional, cast, Union, Any

import elementpath.aliases as ta

from elementpath.protocols import ElementProtocol, DocumentProtocol
from elementpath.datatypes import QName, AnyAtomicType
from elementpath.tdop import MultiLabel
from elementpath.namespaces import XSD_NAMESPACE, XPATH_FUNCTIONS_NAMESPACE, \
    XPATH_MATH_FUNCTIONS_NAMESPACE
from elementpath.helpers import split_function_test, ordinal
from elementpath.sequences import xlist
from elementpath.etree import is_etree_document, is_etree_element
from elementpath.sequence_types import match_sequence_type, is_sequence_type_restriction
from elementpath.xpath_nodes import XPathNode
from elementpath.tree_builders import get_node_tree

from .base import XPathToken
from .tokens import ValueToken


class XPathFunction(XPathToken):
    """
    A token for processing XPath functions.
    """
    __name__: str
    _qname: Optional[QName] = None
    pattern = r'(?<!\$)\b[^\d\W][\w.\-\xb7\u0300-\u036F\u203F\u2040]*' \
              r'(?=\s*(?:\(\:.*\:\))?\s*\((?!\:))'

    sequence_types: ta.SequenceTypesType = ()
    "Sequence types of arguments and of the return value of the function."

    nargs: ta.NargsType = None
    "Number of arguments: a single value or a couple with None that means unbounded."

    context: ta.ContextType = None
    "Dynamic context associated by function reference evaluation or explicitly by a builder."

    def __init__(self, parser: ta.XPathParserType, nargs: Optional[int] = None) -> None:
        super().__init__(parser)
        if isinstance(nargs, int) and nargs != self.nargs:
            if nargs < 0:
                raise self.error('XPST0017', 'number of arguments must be non negative')
            elif self.nargs is None:
                self.nargs = nargs
            elif isinstance(self.nargs, int):
                raise self.error('XPST0017', 'incongruent number of arguments')
            elif self.nargs[0] > nargs or self.nargs[1] is not None and self.nargs[1] < nargs:
                raise self.error('XPST0017', 'incongruent number of arguments')
            else:
                self.nargs = nargs

    def __repr__(self) -> str:
        qname = self.qname
        if qname is None:
            return '<%s object at %#x>' % (self.__class__.__name__, id(self))
        elif not isinstance(self.nargs, int):
            return '<XPathFunction %s at %#x>' % (qname.qname, id(self))
        return '<XPathFunction %s#%r at %#x>' % (qname.qname, self.nargs, id(self))

    def __str__(self) -> str:
        if self.namespace is None:
            return f'{self.symbol!r} {self.label}'
        elif self.namespace == XPATH_FUNCTIONS_NAMESPACE:
            return f"'fn:{self.symbol}' {self.label}"
        else:
            for prefix, uri in self.parser.namespaces.items():
                if uri == self.namespace:
                    return f"'{prefix}:{self.symbol}' {self.label}"
            else:
                return f"'Q{{{self.namespace}}}{self.symbol}' {self.label}"

    def __call__(self, *args: ta.FunctionArgType, context: ta.ContextType = None) -> ta.ValueType:
        self.check_arguments_number(len(args))
        context = copy(self.context or context)

        if self.label == 'partial function':
            for arg, tk in zip(args, filter(lambda x: x.symbol == '?', self)):
                if isinstance(arg, XPathFunction) or not isinstance(arg, XPathToken):
                    tk.value = self.validated_argument(arg, context)
                else:
                    tk.value = arg.evaluate(context)
        else:
            self.clear()
            for arg in args:
                if isinstance(arg, XPathToken):
                    self._items.append(arg)
                else:
                    value = self.validated_argument(arg, context)
                    # Accepts and wraps etree elements/documents, useful for external calls.
                    self._items.append(ValueToken(self.parser, value=value))

            if any(tk.symbol == '?' and not tk for tk in self._items):
                self.to_partial_function()
                return self

        if isinstance(self.label, MultiLabel):
            # Disambiguate multi-label tokens
            if self.namespace == XSD_NAMESPACE and \
                    'constructor function' in self.label.values:
                self.label = 'constructor function'
            else:
                for label in self.label.values:
                    if label.endswith('function'):
                        self.label = label
                        break

        if self.label == 'partial function':
            result = self._partial_evaluate(context)
        else:
            result = self.evaluate(context)

        return self.validated_result(result)

    def check_arguments_number(self, nargs: int) -> None:
        """Check the number of arguments against function arity."""
        if self.nargs is None or self.nargs == nargs:
            pass
        elif isinstance(self.nargs, tuple):
            if nargs < self.nargs[0]:
                raise self.error('XPTY0004', "missing required arguments")
            elif self.nargs[1] is not None and nargs > self.nargs[1]:
                raise self.error('XPTY0004', "too many arguments")
        elif self.nargs > nargs:
            raise self.error('XPTY0004', "missing required arguments")
        else:
            raise self.error('XPTY0004', "too many arguments")

    def validated_argument(self, arg: ta.FunctionArgType,
                           context: ta.ContextType = None) -> ta.ValueType:

        def get_arg_item(item: ta.FunctionArgType) -> ta.ItemType:
            if isinstance(item, (XPathNode, XPathFunction, AnyAtomicType)):
                return item
            elif not is_etree_document(item) and not is_etree_element(item):
                raise self.error('XPTY0004', f"unexpected argument type {type(item)}")
            else:
                return get_node_tree(
                    cast(Union[ElementProtocol, DocumentProtocol], item),
                    namespaces=self.parser.namespaces,
                    uri=self.parser.base_uri,
                    fragment=None
                )

        if context is not None:
            return context.get_value(arg, context.namespaces, self.parser.base_uri, None)
        elif arg is None:
            return []
        elif not isinstance(arg, (list, tuple)):
            return get_arg_item(arg)
        return [get_arg_item(x) for x in arg]

    def validated_result(self, result: ta.ValueType) -> ta.ValueType:
        if isinstance(result, XPathToken) and result.symbol == '?':
            return result
        elif match_sequence_type(result, self.sequence_types[-1], self.parser):
            return result

        result = self.cast_to_primitive_type(result, self.sequence_types[-1])
        if not match_sequence_type(result, self.sequence_types[-1], self.parser):
            msg = "{!r} does not match sequence type {}"
            raise self.error('XPTY0004', msg.format(result, self.sequence_types[-1]))
        return result

    @property
    def source(self) -> str:
        return f"{self.symbol}({', '.join(t.source for t in self)}){self.occurrence}"

    @property
    def qname(self) -> Optional[QName]:
        if self._qname is not None:
            return self._qname
        elif self.symbol == 'function':
            return None
        elif self.label == 'partial function':
            return None
        elif not self.namespace:
            self._qname = QName(None, self.symbol)
        elif self.namespace == XPATH_FUNCTIONS_NAMESPACE:
            self._qname = QName(XPATH_FUNCTIONS_NAMESPACE, 'fn:%s' % self.symbol)
        elif self.namespace == XSD_NAMESPACE:
            self._qname = QName(XSD_NAMESPACE, 'xs:%s' % self.symbol)
        elif self.namespace == XPATH_MATH_FUNCTIONS_NAMESPACE:
            self._qname = QName(XPATH_MATH_FUNCTIONS_NAMESPACE, 'math:%s' % self.symbol)
        else:
            for pfx, uri in self.parser.namespaces.items():
                if uri == self.namespace:
                    self._qname = QName(uri, f'{pfx}:{self.symbol}')
                    break
            else:
                self._qname = QName(self.namespace, self.symbol)

        return self._qname

    @property
    def arity(self) -> int:
        if isinstance(self.nargs, int):
            return self.nargs
        return len(self._items)

    @property
    def min_args(self) -> int:
        if isinstance(self.nargs, int):
            return self.nargs
        elif isinstance(self.nargs, (tuple, list)):
            return self.nargs[0]
        else:
            return 0

    @property
    def max_args(self) -> Optional[int]:
        if isinstance(self.nargs, int):
            return self.nargs
        elif isinstance(self.nargs, (tuple, list)):
            return self.nargs[1]
        else:
            return None

    def is_reference(self) -> int:
        if not isinstance(self.nargs, int):
            return False
        return self.nargs and not len(self._items)

    def nud(self) -> 'XPathFunction':
        if not self.parser.parse_arguments:
            return self

        code = 'XPST0017' if 'function' in self.label else 'XPST0003'
        self.parser.advance('(')
        if self.nargs is None:
            del self._items[:]
            if self.parser.next_token.symbol in (')', '(end)'):
                raise self.error(code, 'at least an argument is required')
            while True:
                self.append(self.parser.expression(5))
                if self.parser.next_token.symbol != ',':
                    break
                self.parser.advance()
        elif self.nargs == 0:
            if self.parser.next_token.symbol != ')':
                if self.parser.next_token.symbol != '(end)':
                    raise self.error(code, '%s has no arguments' % str(self))
                raise self.parser.next_token.wrong_syntax()
            self.parser.advance()
            return self
        else:
            if isinstance(self.nargs, (tuple, list)):
                min_args, max_args = self.nargs
            else:
                min_args = max_args = self.nargs

            k = 0
            while k < min_args:
                if self.parser.next_token.symbol in (')', '(end)'):
                    msg = 'Too few arguments: expected at least %s arguments' % min_args
                    raise self.error('XPST0017', msg if min_args > 1 else msg[:-1])

                self._items[k:] = self.parser.expression(5),
                k += 1
                if k < min_args:
                    if self.parser.next_token.symbol == ')':
                        msg = f'{str(self)}: Too few arguments, expected ' \
                              f'at least {min_args} arguments'
                        raise self.error(code, msg if min_args > 1 else msg[:-1])
                    self.parser.advance(',')

            while max_args is None or k < max_args:
                if self.parser.next_token.symbol == ',':
                    self.parser.advance(',')
                    self._items[k:] = self.parser.expression(5),
                elif k == 0 and self.parser.next_token.symbol != ')':
                    self._items[k:] = self.parser.expression(5),
                else:
                    break  # pragma: no cover
                k += 1

            if self.parser.next_token.symbol == ',':
                msg = 'Too many arguments: expected at most %s arguments' % max_args
                raise self.error(code, msg if max_args != 1 else msg[:-1])

        self.parser.advance(')')
        if any(tk.symbol == '?' and not tk for tk in self._items):
            self.to_partial_function()

        return self

    def match_function_test(self, function_test: ta.SequenceTypesType,
                            as_argument: bool = False) -> bool:
        """
        Match if function signature satisfies the provided *function_test*.
        For default return type is covariant and arguments are contravariant.
        If *as_argument* is `True` the match is inverted.

        References:
          https://www.w3.org/TR/xpath-31/#id-function-test
          https://www.w3.org/TR/xpath-31/#id-sequencetype-subtype
        """
        if isinstance(function_test, (list, tuple)):
            sequence_types = function_test
        else:
            sequence_types = split_function_test(function_test)
        if not sequence_types or not sequence_types[-1]:
            return False
        elif sequence_types[0] == '*':
            return True

        signature = [x for x in self.sequence_types[:self.arity]]
        signature.append(self.sequence_types[-1])

        if len(sequence_types) != len(signature):
            return False

        if as_argument:
            iterator = zip(sequence_types[:-1], signature[:-1])
        else:
            iterator = zip(signature[:-1], sequence_types[:-1])

        # compare sequence types
        for st1, st2 in iterator:
            if not is_sequence_type_restriction(st1, st2):
                return False
        else:
            st1, st2 = sequence_types[-1], signature[-1]
            return is_sequence_type_restriction(st1, st2)

    def to_partial_function(self) -> None:
        """Convert an XPath function to a partial function."""
        nargs = len([tk and not tk for tk in self._items if tk.symbol == '?'])
        assert nargs, "a partial function requires at least a placeholder token"

        if self.label != 'partial function':
            def evaluate(context: ta.ContextType = None) -> 'XPathFunction':
                return self

            def select(context: ta.ContextType = None) -> Iterator['XPathFunction']:
                yield self

            if self.__class__.evaluate is not XPathToken.evaluate:
                setattr(self, '_partial_evaluate', self.evaluate)
            if self.__class__.select is not XPathToken.select:
                setattr(self, '_partial_select', self.select)

            setattr(self, 'evaluate', evaluate)
            setattr(self, 'select', select)

        self._qname = None
        self.label = 'partial function'
        self.nargs = nargs

    def as_function(self) -> Callable[..., Any]:
        """
        Wraps the XPath function instance into a standard function.
        """
        def wrapper(*args: ta.FunctionArgType, context: ta.ContextType = None) -> ta.ValueType:
            return self.__call__(*args, context=context)

        qname = self.qname
        if self.is_reference():
            ref_part = f'#{self.nargs}'
        else:
            ref_part = ''

        if qname is None:
            name = f'<anonymous-function{ref_part}>'
        else:
            name = f'<{qname.qname}{ref_part}>'

        wrapper.__name__ = name
        wrapper.__qualname__ = wrapper.__qualname__[:-7] + name
        return wrapper

    def _partial_evaluate(self, context: ta.ContextType = None) -> ta.ValueType:
        return xlist(self._partial_select(context))

    def _partial_select(self, context: ta.ContextType = None) -> Iterator[ta.ItemType]:
        item = self._partial_evaluate(context)
        if item is not None:
            if isinstance(item, list):
                yield from item
            else:
                if context is not None:
                    context.item = item
                yield item


class ExternalFunction(XPathFunction):
    callback: Callable[..., Any]

    def evaluate(self, context: ta.ContextType = None) -> Any:
        args = []
        for k in range(len(self)):
            try:
                if self.sequence_types[k][-1] in '+*':
                    arg = self[k].evaluate(context)
                else:
                    arg = self.get_argument(context, index=k)
            except IndexError:
                arg = self.get_argument(context, index=k)
            args.append(arg)

        if self.sequence_types:
            for k, (arg, st) in enumerate(zip(args, self.sequence_types), start=1):
                if not match_sequence_type(arg, st, self.parser):
                    msg = f"{ordinal(k)} argument does not match sequence type {st!r}"
                    raise self.error('XPDY0050', msg)

            result = self.callback(*args)
            if not match_sequence_type(result, self.sequence_types[-1], self.parser):
                msg = f"Result does not match sequence type {self.sequence_types[-1]!r}"
                raise self.error('XPDY0050', msg)
            return result

        return self.callback(*args)



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from typing import Any

import elementpath.aliases as ta

from elementpath.exceptions import ElementPathError, MissingContextError
from elementpath.datatypes import AnyAtomicType, ListType, UntypedAtomic
from elementpath.xpath_context import XPathContext, XPathSchemaContext
from .functions import XPathFunction


class XPathConstructor(XPathFunction):
    """
    A token for processing XPath 2.0+ constructors.
    """
    type_class: type[AnyAtomicType | ListType]

    @staticmethod
    def cast(value: Any) -> ta.AtomicType:
        raise NotImplementedError()

    def nud(self) -> 'XPathConstructor':
        if not self.parser.parse_arguments:
            return self

        try:
            self.parser.advance('(')
            self[0:] = self.parser.expression(5),
            if self.parser.next_token.symbol == ',':
                msg = 'Too many arguments: expected at most 1 argument'
                raise self.error('XPST0017', msg)
            self.parser.advance(')')
        except SyntaxError:
            raise self.error('XPST0017') from None
        else:
            if self[0].symbol == '?':
                self.to_partial_function()
            return self

    def evaluate(self, context: XPathContext | None = None) -> ta.OneAtomicOrEmpty:
        if self.context is not None:
            context = self.context

        arg = self.data_value(self.get_argument(context))
        if arg is None:
            return []
        elif arg == '?' and self[0].symbol == '?':
            raise self.error('XPTY0004', "cannot evaluate a partial function")

        try:
            if isinstance(arg, UntypedAtomic):
                return self.cast(arg.value)
            return self.cast(arg)
        except ElementPathError:
            raise
        except (TypeError, ValueError) as err:
            if isinstance(context, XPathSchemaContext):
                return []
            raise self.error('FORG0001', err) from None


class SchemaConstructor(XPathFunction):
    """
    A token for processing XPath 2.0+ constructors defined by schema atomic types.
    """
    type_class: type[AnyAtomicType | ListType]

    @staticmethod
    def cast(value: Any) -> ta.AtomicType:
        raise NotImplementedError()

    def nud(self) -> 'SchemaConstructor':
        self.parser.advance('(')
        self[0:] = self.parser.expression(5),
        self.parser.advance(')')

        try:
            self.evaluate()  # for static context evaluation
        except MissingContextError:
            pass
        return self

    def evaluate(self, context: ta.ContextType = None) -> ta.OneAtomicOrEmpty:
        arg = self.get_argument(context)
        if arg is None or self.parser.schema is None:
            return []

        value = self.string_value(arg)
        try:
            return self.parser.schema.cast_as(value, self.name)
        except (TypeError, ValueError) as err:
            if isinstance(context, XPathSchemaContext):
                return []
            raise self.error('FORG0001', err)
